############
#
# Copyright (c) 2024-2026 Maxim Yudayev and KU Leuven eMedia Lab
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Created 2024-2025 for the KU Leuven AidWear, AidFOG, and RevalExo projects
# by Maxim Yudayev [https://yudayev.com].
#
# ############

import time
from typing import Callable
import numpy as np


def estimate_transmission_delay(ping_fn: Callable, num_samples: int = 100) -> float:
    """Estimates the mean transmission delay of the provided "ping" function.

    Used with the custom provided `ping_fn` to measure the one-way delay to use for
    offsetting sensor data streams from the `toa_s` to obtain true sample time.

    Args:
        ping_fn (Callable): User's function that wraps "ping" like functionality to the specific device.
        num_samples (int, optional): Number of round-trip transmission to perform to average over. Defaults to `100`.

    Returns:
        float: Estimated mean delay for samples from the device.
    """
    transmit_delays_s: list[float] = []
    for i in range(num_samples):
        local_time_before = time.time()
        ping_fn()
        local_time_after = time.time()
        # Assume symmetric delays.
        transmit_delays_s.append((local_time_after - local_time_before) / 2.0)
    # TODO: remove outliers before averaging.
    delays = np.array(transmit_delays_s)
    return float(np.mean(delays))
